<?php
namespace Ep\Http\Middleware;

use GuzzleHttp\Promise\FulfilledPromise;
use GuzzleHttp\Promise\Promise;
use function GuzzleHttp\Psr7\parse_response;
use function GuzzleHttp\Psr7\str;
use Psr\Cache\CacheItemPoolInterface;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;

/**
 * Kešira HTTP responsove.
 *
 * @package Ep\Http\Middleware
 * @author Elektronika Perhoč <web@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/core
 */
class Cache
{
    const HEADER_CACHE_INFO  = 'X-Http-Cache';
    const HEADER_CACHE_HIT   = 'HIT';
    const HEADER_CACHE_MISS  = 'MISS';

    /** @var CacheItemPoolInterface */
    private $cache;

    /**
     * Cache constructor.
     *
     * @param CacheItemPoolInterface $cache
     */
    public function __construct(CacheItemPoolInterface $cache)
    {
        $this->cache = $cache;
    }

    /**
     * Traži request u cacheu.
     *
     * @param callable $handler
     *
     * @return \Closure
     */
    public function __invoke(callable $handler)
    {
        return function(RequestInterface $request, array $options) use($handler) {
            $cacheKey = $this->generateCacheKey($request);
            $cacheItem = $this->cache->getItem($cacheKey);

            if ($cacheItem->isHit()) {
                $response = parse_response($cacheItem->get());

                return new FulfilledPromise(
                    $response->withAddedHeader(self::HEADER_CACHE_INFO, self::HEADER_CACHE_HIT)
                );
            }

            /** @var Promise $promise */
            $promise = $handler($request, $options);

            return $promise->then(function(ResponseInterface $response) use ($cacheItem) {
                if ($response->getStatusCode() < 300) {
                    // neki apiji i za greške vraćaju 200 status code pa moramo analizirati sadržaj
                    if ('application/json' === $response->getHeaderLine('Content-Type')) {
                        $json = json_decode($response->getBody()->getContents(), true);

                        if (isset($json['status']) && $json['status'] >= 300) {
                            return $response->withAddedHeader(self::HEADER_CACHE_INFO, self::HEADER_CACHE_MISS);
                        }
                    }

                    $cacheItem->set(str($response));
                    $cacheControlHeader = $response->getHeaderLine('Cache-Control');

                    if (
                        $response->hasHeader('Cache-Control') && preg_match('/max-age=(\d+)/', $cacheControlHeader, $matches)
                    ) {
                        $expiresAfter = intval($matches[1]);
                        $cacheItem->expiresAfter($expiresAfter);
                    }

                    if (!preg_match('/no-cache/', $cacheControlHeader)) {
                        $this->cache->save($cacheItem);
                    }
                }

                $response->getBody()->rewind();
                return $response->withAddedHeader(self::HEADER_CACHE_INFO, self::HEADER_CACHE_MISS);
            });
        };
    }

    /**
     * Generira cache ključ za dati request.
     *
     * @param RequestInterface $request
     *
     * @return string
     */
    private function generateCacheKey(RequestInterface $request)
    {
        $cacheKey = [];
        $cacheKey[] = $request->getMethod();
        $cacheKey[] = $request->getUri();
        $cacheKey[] = $request->getBody()->getContents();

        return hash('sha512', implode('', $cacheKey));
    }
}